import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


lexnlp_base_path = os.path.abspath(os.path.dirname(__file__) + '/../../../')

lexnlp_test_path = os.path.join(lexnlp_base_path, 'test_data/')
