import os

__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


ENV_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
ENV_EN_DATA_DIRECTORY = os.path.join(ENV_DIRECTORY, 'en/data')
