__author__ = "ContraxSuite, LLC; LexPredict, LLC"
__copyright__ = "Copyright 2015-2019, ContraxSuite, LLC"
__license__ = "https://github.com/LexPredict/lexpredict-lexnlp/blob/master/LICENSE"
__version__ = "1.4.0"
__maintainer__ = "LexPredict, LLC"
__email__ = "support@contraxsuite.com"


class EsLanguageTokens:
    """
    Spanish parts of speech, used in a number of parsing methods
    """
    abbreviations = {'nr.', 'abs.', 'no.', 'act.', 'inc.', 'p.'}
    articles = ['el', 'la', 'los', 'las']
    conjunctions = ['und', 'oder']
